/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <sstream>
#include <unistd.h>
#include <list>

#include "log.h"
#include "database.h"
#include "create_database.h"
#include "config_manager.h"
#include "dsmapi-wrapper.h"
#include "hwdesc.h"
#include "storage.h"
#include "util.h"
#include "exceptions.h"

// quote
#define Q(data) "\"" << data << "\""

#define TASKDATA_TRANSMISSION_QUALITY Q("taskdata_test_transmission_quality")
#define TASKDATA_DEVICE_PRESENCE Q("taskdata_test_device_presence")
#define TASKDATA_STATISTIC_DEVICE Q("taskdata_device_statistic")
#define TASKDATA_STATISTIC_METER Q("taskdata_meter_statistic")

namespace dSA
{

Database* Database::m_instance = NULL;

Database* Database::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new Database();
        if (m_instance == NULL)
        {
            throw _Exception("could not create Database class");
        }
    }

    return m_instance;
}

Database* Database::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("Database not initialized");
    }

    return m_instance;
}

Database::Database()
{
    ConfigManager *cfg = ConfigManager::getInstance();

    bool exists = fileutil::exists(cfg->getOption(CFG_DATABASE_FILE));

    m_sqlite3 = new SQLite3(cfg->getOption(CFG_DATABASE_FILE));
    if (m_sqlite3 == NULL)
    {
        throw _Exception("Failed to allocate database");
    }
    if (!exists)
    {
        try
        {
            FILE_LOG(logINFO) << "Creating database version 2";
            m_sqlite3->exec(CREATE_DB_V1);
            m_sqlite3->exec(CREATE_DB_V2);
            m_sqlite3->exec(DB_STATS_V1);
        }
        catch (Exception ex)
        {
            delete m_sqlite3;
            throw _Exception("Could not create database, " + ex.getMessage());
        }
    }

    try
    {
        m_version = util::str2int(getConfig("db_version"));
    }
    catch (SQLite3Exception sx)
    {
        delete m_sqlite3;
        throw _SQLite3Exception("Could not determine database layout "
                                "version: " + sx.getMessage(),
                                sx.getErrorCode());
    }
    catch (Exception ex)
    {
        delete m_sqlite3;
        throw _Exception("Could not determine database layout version: " +
                         ex.getMessage());
    }

    try
    {
        if (m_version == 1)
        {
            FILE_LOG(logINFO) << "Updating database layout from version " <<
                    util::str(m_version) << " to 2";
            m_sqlite3->exec(CREATE_DB_V2);
            m_sqlite3->exec(DB_STATS_V1);
        }
    }
    catch (SQLite3Exception sx)
    {
        delete m_sqlite3;
        throw _SQLite3Exception("Could not update database layout: " +
                                sx.getMessage(), sx.getErrorCode());
    }
    catch (Exception ex)
    {
        delete m_sqlite3;
        throw _Exception("Could not update database layout: " + ex.getMessage());
    }
}

/// \brief Return size of database file on disk
size_t Database::getFileSize() const
{
    std::string filename = ConfigManager::getInstance()->getOption(CFG_DATABASE_FILE);
    return fileutil::fsize(filename);
}

/// \brief Shrink database file size by running the VACUUM; command
void Database::executeVacuum() const
{
    std::ostringstream qstr;
    qstr << "VACUUM;";

    FILE_LOG(logDEBUG) << "Database compression started";
    m_sqlite3->exec(qstr.str());
    FILE_LOG(logDEBUG) << "Database compression finished";
}

// internal settings are saved in a table as key:value pairs
std::string Database::getConfig(std::string key)
{
    std::ostringstream qstr;

    qstr << "SELECT " << Q("value") << " FROM " << Q("dsa_config");
    qstr << " WHERE " << Q("key") << " = ";
    qstr << m_sqlite3->escape(key);
    qstr << " LIMIT 1;";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    // we expect exactly one row (or none if the key does not exist)
    if (q->size() <  1)
    {
        throw _Exception("setting " + key + " was not found in the database");
    }

    // in response we should get exactly one row, containing the value
    boost::shared_ptr<SQLite3::row_result> row = q->at(0);

    if (row->size() < 1)
    {
        throw _Exception("setting " + key + " was not found in the database");
    }

    boost::shared_ptr<SQLite3::cell> cell = row->at(0);
    return cell->data;
}

void Database::setConfig(std::string key, std::string value)
{
    std::ostringstream qstr;

    qstr << "INSERT OR REPLACE INTO " << Q("dsa_config") << " (" <<  Q("key");
    qstr << ", " << Q("value") << ") VALUES (" << m_sqlite3->escape(key);
    qstr << ", " << m_sqlite3->escape(value) << ")";
    m_sqlite3->exec(qstr.str());
}

long long int Database::saveTask(boost::shared_ptr<const dSLongTask> task)
{
    std::ostringstream qstr;

    qstr << "INSERT INTO " << Q("tasks") << " (" << Q("status") << ",";
    qstr << Q("queued") << "," << Q("type") << "," << Q("parameters") << ") ";
    qstr << "VALUES (" << task->getTaskStatus() << ",";
    qstr << "strftime('%s', 'now'),"<< task->getTaskType() << ",";
    qstr << m_sqlite3->escape(task->getParameters()) << ");";
    return m_sqlite3->execAndGetRowId(qstr.str());
}

void Database::removeTask(long long int task_id)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("tasks") << " WHERE " << Q("task_id") << " = ";
    qstr << task_id << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::removeTasksByStatus(task_status status)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("tasks") << " WHERE " << Q("status") << " = ";
    qstr << status << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::updateTaskStatus(long long int task_id, task_status status)
{
    std::ostringstream qstr;

    qstr << "UPDATE " << Q("tasks") << " SET " << Q("status") << " = ";
    qstr << status;
    if (status == Running)
    {
        qstr << "," << Q("started") << " = " << "strftime('%s', 'now')";
    }
    else if (status == Finished)
    {
        qstr << "," << Q("finished") << " = " << "strftime('%s', 'now')";
    }

    qstr << " WHERE " << Q("task_id") << " = " << task_id;
    qstr << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::updateTaskStatus(task_status old_status, task_status new_status)
{
    std::ostringstream qstr;
    qstr << "UPDATE " << Q("tasks") << " SET " << Q("status") << " = ";
    qstr << new_status;
    qstr << " WHERE " << Q("status") << " = " << old_status << ";";
    m_sqlite3->exec(qstr.str());
}

task_type Database::getTaskType(long long int task_id)
{
    std::ostringstream qstr;
    qstr << "SELECT " << Q("type") << " FROM " << Q("tasks") << " WHERE ";
    qstr << Q("task_id") << " = " << task_id << " LIMIT 1;";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("Task with id " + util::str(task_id) + " not found");
    }

    boost::shared_ptr<SQLite3::row_result> rr = q->at(0);
    for (size_t i = 0; i < rr->size(); i++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(i);
        if (cell->name == "type")
        {
            return (task_type)util::str2int(cell->data);
        }
    }

    throw _Exception("Could not find task type information");
}

boost::shared_ptr<dSLongTask> Database::getTask(long long int task_id)
{
    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("tasks") << " WHERE ";
    qstr << Q("task_id") << " = " << task_id << " LIMIT 1;";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("Task with id " + util::str(task_id) + " not found");
    }

    return createTaskFromRow(q->at(0));
}

boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > >
                                 Database::getTasksByStatus(task_status status)
{
    boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > >
        tasks(new std::list<boost::shared_ptr<dSLongTask> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << Q("tasks") << " WHERE " << Q("status") << " = ";
    qstr << status << ";";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            tasks->push_back(createTaskFromRow(q->at(i)));
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) <<
                    "Caught exception while parsing saved tasks: " <<
                    ex.getMessage();
            continue;
        }
    }

    return tasks;
}

std::pair<std::string,std::string> Database::getStatisticDescription(statistic_field_t index)
{
    std::ostringstream qstr;
    qstr << "SELECT * FROM " << Q("dsa_statistic") << " WHERE " << Q("stat_id") << " = ";
    qstr << util::str(index) << ";";

    boost::shared_ptr<SQLite3::query_result> q;
    try
    {
        q = m_sqlite3->query(qstr.str());
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) <<
                "Loading statistic description table for index: " <<
                ex.getMessage();
        throw _Exception("Loading data for field index " + util::str(index) + " failed");
    }

    // in response we should get exactly one row, containing the value
    if (q->size() != 1)
    {
        throw _Exception("Loading data for field index " + util::str(index) + " failed");
    }
    boost::shared_ptr<SQLite3::row_result> row = q->at(0);
    boost::shared_ptr<SQLite3::cell> rname = row->at(1);
    boost::shared_ptr<SQLite3::cell> rdesc = row->at(2);

    std::pair<std::string,std::string> result(rname->data, rdesc->data);
    return result;
}

boost::shared_ptr<std::list<std::pair<statistic_field_t,std::string> > >
    Database::getStatisticTableNames(int mode)
{
    boost::shared_ptr<std::list<std::pair<statistic_field_t,std::string> > >
        result(new std::list<std::pair<statistic_field_t,std::string> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << Q("dsa_statistic");
    qstr << ";";

    boost::shared_ptr<SQLite3::query_result> q;
    try
    {
        q = m_sqlite3->query(qstr.str());
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) <<
                "Loading statistic description table: " <<
                ex.getMessage();
        throw _Exception("Loading statistic table data: " + ex.getMessage());
    }

    for (size_t i = 0; i < q->size(); i++)
    {
        boost::shared_ptr<SQLite3::row_result> rr = q->at(i);
        try
        {
            switch (mode)
            {
                case 0:
                {
                    result->push_back(std::make_pair(
                            (statistic_field_t) util::str2int(rr->at(0)->data),
                            rr->at(1)->data)
                    );
                }
                case 1:
                {
                    result->push_back(std::make_pair(
                            (statistic_field_t) util::str2int(rr->at(0)->data),
                            rr->at(2)->data)
                    );
                }
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) <<
                    "Parsing statistic table rows: " <<
                    ex.getMessage();
            continue;
        }
    }

    return result;
}

void Database::saveTransmissionQuality(long long int task_id,
                                       std::string device_dsid,
                                       uint16_t upstream, uint8_t downstream,
                                       int error_code)
{
    std::ostringstream qstr;

    qstr << "INSERT INTO " << TASKDATA_TRANSMISSION_QUALITY << " (";
    qstr << Q("dsid") << ",";
    qstr << Q("upstream") << ",";
    qstr << Q("downstream") << ",";
    qstr << Q("timestamp") << ",";
    qstr << Q("errorcode") << ",";
    qstr << Q("task_id");
    qstr << ") ";

    qstr << "VALUES (";
    qstr << m_sqlite3->escape(device_dsid) << ",";
    qstr << upstream << ",";
    qstr << (int)downstream << ",";
    qstr << "strftime('%s', 'now'),";
    qstr << error_code << ",";
    qstr << task_id;
    qstr << ");";

    m_sqlite3->exec(qstr.str());
}

void Database::saveDevicePresence(long long int task_id,
        std::string device_dsid, std::string dsm_dsid, uint16_t device_id,
        uint8_t resultcode)
{
    std::ostringstream qstr;

    qstr << "INSERT INTO " << TASKDATA_DEVICE_PRESENCE << " (";
    qstr << Q("dsid") << ",";
    qstr << Q("dsmdsid") << ",";
    qstr << Q("deviceid") << ",";
    qstr << Q("resultcode") << ",";
    qstr << Q("timestamp") << ",";
    qstr << Q("task_id");
    qstr << ") ";

    qstr << "VALUES (";
    qstr << m_sqlite3->escape(device_dsid) << ",";
    qstr << m_sqlite3->escape(dsm_dsid) << ",";
    qstr << (int) device_id << ",";
    qstr << (int) resultcode << ",";
    qstr << "strftime('%s', 'now'),";
    qstr << task_id;
    qstr << ");";

    m_sqlite3->exec(qstr.str());
}

void Database::updateDevicePresenceStatus(long long int task_id,
        std::string dsm_dsid, uint16_t device_id, uint8_t resultcode)
{
    std::ostringstream qstr;

    qstr << "UPDATE " << TASKDATA_DEVICE_PRESENCE << " SET ";
    qstr << Q("resultcode") << " = " << (int) resultcode << ", ";
    qstr << Q("timestamp") << " = " << "strftime('%s', 'now') ";
    qstr << "WHERE (";
    qstr << Q("task_id") << " = " << task_id;
    qstr << " AND " << Q("deviceid") << " = " << (int) device_id;
    qstr << " AND " << Q("dsmdsid") << " = " << m_sqlite3->escape(dsm_dsid) << ");";

    m_sqlite3->exec(qstr.str());
}

void Database::saveStatisticDevice(long long int task_id,
        dsid_t device_dsid, dsid_t dsm_dsid, uint16_t device_id,
        statistic_field_t index, long long int value,
        uint8_t resultcode)
{
    std::ostringstream qstr;

    qstr << "INSERT INTO " << TASKDATA_STATISTIC_DEVICE << " (";
    qstr << Q("dsid") << ",";
    qstr << Q("dsmdsid") << ",";
    qstr << Q("deviceid") << ",";
    qstr << Q("value") << ",";
    qstr << Q("timestamp") << ",";
    qstr << Q("errorcode") << ",";
    qstr << Q("stat_id") << ",";
    qstr << Q("task_id");
    qstr << ") ";

    qstr << "VALUES (";
    qstr << m_sqlite3->escape(util::str(device_dsid)) << ",";
    qstr << m_sqlite3->escape(util::str(dsm_dsid)) << ",";
    qstr << (int) device_id << ",";
    qstr << value << ",";
    qstr << "strftime('%s', 'now'),";
    qstr << (int) resultcode << ",";
    qstr << (int) index << ",";
    qstr << task_id;
    qstr << ");";

    m_sqlite3->exec(qstr.str());
}

void Database::saveStatisticDevice(long long int task_id,
        dsid_t device_dsid, dsid_t dsm_dsid, uint16_t device_id,
        boost::shared_ptr<std::list<boost::shared_ptr<StatisticValue> > > svalues)
{
    std::ostringstream qstr;
    std::string d_dsid(m_sqlite3->escape(util::str(device_dsid)));
    std::string m_dsid(m_sqlite3->escape(util::str(dsm_dsid)));

    qstr << "BEGIN TRANSACTION;";

    for (std::list<boost::shared_ptr<StatisticValue> >::iterator it = svalues->begin();
         it != svalues->end();
         it++)
    {
        qstr << "INSERT INTO " << TASKDATA_STATISTIC_DEVICE << " (";
        qstr << Q("dsid") << ",";
        qstr << Q("dsmdsid") << ",";
        qstr << Q("deviceid") << ",";
        qstr << Q("value") << ",";
        qstr << Q("timestamp") << ",";
        qstr << Q("errorcode") << ",";
        qstr << Q("stat_id") << ",";
        qstr << Q("task_id");
        qstr << ") ";

        qstr << "VALUES (";
        qstr << d_dsid << ",";
        qstr << m_dsid << ",";
        qstr << (int) device_id << ",";
        qstr << (*it)->m_value << ",";
        qstr << (*it)->m_time << ",";
        qstr << (*it)->m_errorcode << ",";
        qstr << (int) (*it)->m_field << ",";
        qstr << task_id;
        qstr << ");";
    }

    qstr << "COMMIT;";
    m_sqlite3->exec(qstr.str());
}

void Database::saveStatisticMeter(long long int task_id, dsid_t dsm_dsid,
        boost::shared_ptr<std::list<boost::shared_ptr<StatisticValue> > > svalues)
{
    std::ostringstream qstr;
    std::string m_dsid(m_sqlite3->escape(util::str(dsm_dsid)));

    qstr << "BEGIN TRANSACTION;";

    for (std::list<boost::shared_ptr<StatisticValue> >::iterator it = svalues->begin();
         it != svalues->end();
         it++)
    {
        qstr << "INSERT INTO " << TASKDATA_STATISTIC_METER << " (";
        qstr << Q("dsmdsid") << ",";
        qstr << Q("value") << ",";
        qstr << Q("timestamp") << ",";
        qstr << Q("errorcode") << ",";
        qstr << Q("stat_id") << ",";
        qstr << Q("task_id");
        qstr << ") ";

        qstr << "VALUES (";
        qstr << m_dsid << ",";
        qstr << (*it)->m_value << ",";
        qstr << (*it)->m_time << ",";
        qstr << (*it)->m_errorcode << ",";
        qstr << (int) (*it)->m_field << ",";
        qstr << task_id;
        qstr << ");";
    }

    qstr << "COMMIT;";
    m_sqlite3->exec(qstr.str());
}

void Database::saveUpstreamMetric(
        boost::shared_ptr<std::list<boost::shared_ptr<dSMApiSelectiveListener::CBData> > > cbdatalist)
{
    std::ostringstream qstr;
    qstr << "BEGIN TRANSACTION;";

    for (std::list<boost::shared_ptr<dSMApiSelectiveListener::CBData> >::iterator it = cbdatalist->begin();
            it != cbdatalist->end();
            it++)
    {

        switch((*it)->getType())
        {
            default:
            break;

            case dSMApiSelectiveListener::CommUpstreamShort:
            {
                boost::shared_ptr<dSMApiSelectiveListener::CommUpstreamShortData> data =
                        boost::dynamic_pointer_cast<dSMApiSelectiveListener::CommUpstreamShortData> (*it);

                dsid_t devid;
                try
                {
                    boost::shared_ptr<const Device> dsd =
                        Storage::getInstance()->getDevice(
                                data->getDeviceID(),
                                data->getDSMSourceDSID());
                    devid = dsd->getDSID();
                }
                catch (Exception &ex)
                {
                    SetNullId(devid);
                }

                qstr << "INSERT INTO \"taskdata_upstream_metric\" ( \"dsid\",\"dsmdsid\","
                        "\"deviceid\",\"quality\",\"polarity\",\"crosstalk\","
                        "\"packettype\",\"timestamp\" )";

                qstr << "VALUES (";
                qstr << "\"" << util::str(devid) << "\",";
                qstr << "\"" << util::str(data->getDSMSourceDSID()) << "\",";
                qstr << (int) data->getDeviceID() << ",";
                qstr << (int) data->getQuality() << ",";
                qstr << (int) data->getFlags() << ",";
                qstr << (int) data->getCrosstalkFlags() << ",";
                qstr << (int) 1 << ",";
                qstr << data->getTimestamp() << ");";
            }
            break;

            case dSMApiSelectiveListener::CommUpstreamLong:
            {
                boost::shared_ptr<dSMApiSelectiveListener::CommUpstreamLongData> data =
                        boost::dynamic_pointer_cast<dSMApiSelectiveListener::CommUpstreamLongData> (*it);
                boost::shared_ptr<const Device> dsd;

                dsid_t devid;
                int busid = 0;
                try
                {
                    if (data->getMainQualifier() == 15)
                    {
                        // NOTE: this expansion works only for release-1 aizo components
                        if (DsmApiExpandDeviceDSID(0x0001, data->getData(), &devid) == 0)
                        {
                        }
                    }
                    else
                    {
                        dsd = Storage::getInstance()->getDevice(
                                data->getDeviceID(),
                                data->getDSMSourceDSID());
                        devid = dsd->getDSID();
                        busid = data->getDeviceID();
                    }
                }
                catch (Exception &ex)
                {
                    SetNullId(devid);
                }

                qstr << "INSERT INTO \"taskdata_upstream_metric\" ( \"dsid\",\"dsmdsid\","
                        "\"deviceid\",\"quality\",\"polarity\",\"crosstalk\","
                        "\"packettype\",\"timestamp\" )";

                qstr << "VALUES (";
                qstr << "\"" << util::str(devid) << "\",";
                qstr << "\"" << util::str(data->getDSMSourceDSID()) << "\",";
                qstr << (int) busid << ",";
                qstr << (int) data->getQuality() << ",";
                qstr << (int) data->getFlags() << ",";
                qstr << (int) data->getCrosstalkFlags() << ",";
                qstr << (int) 2 << ",";
                qstr << data->getTimestamp() << ");";
            }
            break;

        }

        (*it).reset();

    }

    qstr << "COMMIT;";
    m_sqlite3->exec(qstr.str());
}

void Database::removeUpstreamMetricByTime(time_t threshold)
{
    time_t tlimit = time(NULL) - threshold;
    std::ostringstream qstr;

    qstr << "DELETE FROM \"taskdata_upstream_metric\" WHERE " <<
            Q("timestamp") << " < " << tlimit << ";";

    m_sqlite3->query(qstr.str());
}

boost::shared_ptr<Database::task_result_summary>
                          Database::getTaskResultSummary(long long int task_id)
{
    boost::shared_ptr<Database::task_result_summary> result(
            new Database::task_result_summary);

    boost::shared_ptr<result_data> data(new result_data);

    result->task = getTask(task_id);
    result->data = data;

    std::ostringstream qstr;

    boost::shared_ptr<SQLite3::query_result> q;

    if (result->task->getTaskType() == TestTransmissionQuality)
    {
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
        qstr << "WHERE " << Q("errorcode") << " != 0 AND ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["errors"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
        qstr << "WHERE " << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["tests"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT COUNT(DISTINCT " << Q("dsid") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["devices"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MAX( " << Q("upstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["maxUpstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MIN( " << Q("upstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["minUpstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT AVG( " << Q("upstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["avgUpstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MAX( " << Q("downstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["maxDownstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT MIN( " << Q("downstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["minDownstream"] = singleValueFromQuery(q);

        qstr.clear();
        qstr.str("");
        qstr << "SELECT AVG( " << Q("downstream") << ") FROM ";
        qstr << TASKDATA_TRANSMISSION_QUALITY << " " << "WHERE ";
        qstr << Q("task_id") << " = " << task_id;
        qstr << " AND " << Q("errorcode") << " = 0" << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["avgDownstream"] = singleValueFromQuery(q);
    }

    else if (result->task->getTaskType() == TestDevicePresence)
    {
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_DEVICE_PRESENCE << " ";
        qstr << "WHERE " << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["tests"] = singleValueFromQuery(q);

        qstr.str("");
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_DEVICE_PRESENCE << " ";
        qstr << "WHERE " << Q("resultcode") << " == 4 AND ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["replies"] = singleValueFromQuery(q);

        qstr.str("");
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_DEVICE_PRESENCE << " ";
        qstr << "WHERE " << Q("resultcode") << " == 5 AND ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["inactive"] = singleValueFromQuery(q);

        qstr.str("");
        qstr << "SELECT COUNT(*) FROM " << TASKDATA_DEVICE_PRESENCE << " ";
        qstr << "WHERE " << Q("resultcode") << " == 6 AND ";
        qstr << Q("task_id") << " = " << task_id << ";";
        q = m_sqlite3->query(qstr.str());
        (*data)["active"] = singleValueFromQuery(q);
    }

    else
    {
        throw _Exception("could not load task data, unknown task");
    }

    return result;
}

boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                 Database::loadTasks(long long int limit,
                                                     long long int offset)
{
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
        tasks(new std::list<boost::shared_ptr<const dSLongTaskInfo> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << Q("tasks");
    if (limit == 0)
    {
        limit = -1;
    }

    qstr << " LIMIT " << limit;

    if (offset > 0)
    {
        qstr << " OFFSET " << offset;
    }

    qstr << ";";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            tasks->push_back(createTaskFromRow(q->at(i)));
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) <<
                    "Caught exception while parsing saved tasks: " <<
                    ex.getMessage();
            continue;
        }
    }

    return tasks;
}
boost::shared_ptr<dSLongTask> Database::createTaskFromRow(
                                     boost::shared_ptr<SQLite3::row_result> rr)
{
    task_type tt = UnknownTask;
    task_status ts = UnknownStatus;
    webutil::query params;
    long long int task_id = 0;
    time_t started = 0;
    time_t finished = 0;
    time_t queued = 0;

    for (size_t j = 0; j < rr->size(); j++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(j);
        if (cell->name == "type")
        {
            tt = (task_type)util::str2int(cell->data);
        }
        else if (cell->name == "status")
        {
            ts = (task_status)util::str2int(cell->data);
        }
        else if (cell->name == "parameters")
        {
            webutil::parse_query(cell->data, params);
        }
        else if (cell->name == "task_id")
        {
            task_id = util::str2llint(cell->data);
        }
        else if (cell->name == "started")
        {
            if (!cell->data.empty())
            {
                started = util::str2time_t(cell->data);
            }
        }
        else if (cell->name == "queued")
        {
            if (!cell->data.empty())
            {
                queued = util::str2time_t(cell->data);
            }
        }
        else if (cell->name == "finished")
        {
            if (!cell->data.empty())
            {
                finished = util::str2time_t(cell->data);
            }
        }
    }

    boost::shared_ptr<dSLongTask> task;
    switch (tt)
    {
        case TestTransmissionQuality:
        {
            if (params.size() < 2)
            {
                throw _Exception("not enough parameters loaded for "
                        "test transmission quality task");
            }
            task = boost::shared_ptr<dSLongTask> (new TestTranmissionQualityTask(
                    util::str2int(params["checksPerDevice"]),
                    util::str2dsid(params["dsm"])));
        }
        break;

        case TestDevicePresence:
        {
            if (params.size() < 2)
            {
                throw _Exception("not enough parameters loaded for "
                        "test device presence task");
            }
            task = boost::shared_ptr<dSLongTask> (new TestDevicePresenceTask(
                    util::str2dsid(params["dsm"]),
                    util::str2int(params["interval"])));
        }
        break;

        case TestDeviceStatistic:
        {
            if (params.size() < 2)
            {
                throw _Exception("not enough parameters loaded for "
                        "test device presence task");
            }
            task = boost::shared_ptr<dSLongTask> (new TestDeviceStatisticTask(
                    util::str2dsid(params["dsm"]),
                    util::str2dsid(params["dsid"])));
        }
        break;

        case TestSystemMonitor:
        {
            if (params.size() < 2)
            {
                throw _Exception("not enough parameters loaded for "
                        "test device presence task");
            }
            task = boost::shared_ptr<dSLongTask> (new TestSystemMonitorTask(
                    util::str2int(params["profile"]),
                    util::str2int(params["duration"])));
        }
        break;

        default:
        {
            throw _Exception("failed to identify task, unknown task type");
        }
    }

    task->setTaskStatus(ts);
    task->setTaskID(task_id);
    task->setStartedTime(started);
    task->setFinishedTime(finished);
    task->setQueuedTime(queued);
    return task;
}

boost::shared_ptr<Database::field> Database::singleValueFromQuery(
                                    boost::shared_ptr<SQLite3::query_result> q)
{
    if (q->size() == 0)
    {
        throw _Exception("failed to retrieve result");
    }

    boost::shared_ptr<SQLite3::row_result> rr = q->at(0);

    if (rr->size() == 0)
    {
        throw _Exception("failed to retrieve result");
    }
    boost::shared_ptr<SQLite3::cell> cell = rr->at(0);

    boost::shared_ptr<Database::field> d(
            new Database::field(cell->data, cell->type));
    return d;
}

long long int Database::transmissionQualityResultCount(long long int task_id)
{
    std::ostringstream qstr;
    qstr << "SELECT COUNT(*) FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
    qstr << "WHERE " << Q("task_id") << " = " << task_id << ";";

    boost::shared_ptr<SQLite3::query_result> q;
    q = m_sqlite3->query(qstr.str());
    boost::shared_ptr<Database::field> f = singleValueFromQuery(q);

    if (f->second != SQLITE_INTEGER)
    {
        throw _Exception("unexpected datatype");
    }
    return util::str2llint(f->first);
}


boost::shared_ptr<Database::transmission_quality_device_record>
    Database::TQDevRecordFromRow(boost::shared_ptr<SQLite3::row_result> rr)
{
    boost::shared_ptr<transmission_quality_device_record>
        rec(new Database::transmission_quality_device_record);

    for (size_t j = 0; j < rr->size(); j++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(j);
        if (cell->name == "dsid")
        {
            rec->dsid = cell->data;
        }
        else if (cell->name == "upstream")
        {
            rec->upstream = util::str2int(cell->data);
        }
        else if (cell->name == "downstream")
        {
            rec->downstream = util::str2int(cell->data);
        }
        else if (cell->name == "timestamp")
        {
            rec->timestamp = util::str2time_t(cell->data);
        }
        else if (cell->name == "errorcode")
        {
            rec->errorcode = util::str2int(cell->data);
        }
    }

    return rec;
}

boost::shared_ptr<std::list<
    boost::shared_ptr<const Database::transmission_quality_device_record> > >
            Database::getTransmissionQualityResults(long long int task_id,
                                                    long long int limit,
                                                    long long int offset)
{
    boost::shared_ptr<std::list<
        boost::shared_ptr<
            const Database::transmission_quality_device_record> > >
                ret(new std::list<
                    boost::shared_ptr<
                        const Database::transmission_quality_device_record> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << TASKDATA_TRANSMISSION_QUALITY << " ";
    qstr << "WHERE " << Q("task_id") << " = " << task_id;
    if (limit == 0)
    {
        limit = -1;
    }

    qstr << " LIMIT " << limit;

    if (offset > 0)
    {
        qstr << " OFFSET " << offset;
    }

    qstr << ";";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            ret->push_back(TQDevRecordFromRow(q->at(i)));
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) <<
                    "Caught exception while parsing test results: " <<
                    ex.getMessage();
            continue;
        }
    }

    return ret;
}

boost::shared_ptr<std::list<
    boost::shared_ptr<const Database::device_presence_record> > >
            Database::getDevicePresenceResults(long long int task_id,
                    long long int limit,
                    long long int offset)
{
    boost::shared_ptr<std::list<
        boost::shared_ptr<
            const Database::device_presence_record> > >
                ret(new std::list<
                    boost::shared_ptr<
                        const Database::device_presence_record> >);

    std::ostringstream qstr;
    qstr << "SELECT * FROM " << TASKDATA_DEVICE_PRESENCE << " ";
    qstr << "WHERE " << Q("task_id") << " = " << task_id;
    if (limit == 0)
    {
        limit = -1;
    }

    qstr << " LIMIT " << limit;

    if (offset > 0)
    {
        qstr << " OFFSET " << offset;
    }

    qstr << ";";

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        try
        {
            boost::shared_ptr<SQLite3::row_result> rr = q->at(i);
            boost::shared_ptr<device_presence_record>
                rec(new Database::device_presence_record);

            for (size_t j = 0; j < rr->size(); j++)
            {
                boost::shared_ptr<SQLite3::cell> cell = rr->at(j);
                if (cell->name == "dsid")
                {
                    rec->dsid = cell->data;
                }
                if (cell->name == "dsmdsid")
                {
                    rec->dsmdid = cell->data;
                }
                else if (cell->name == "deviceid")
                {
                    rec->device_id = util::str2int(cell->data);
                }
                else if (cell->name == "resultcode")
                {
                    rec->resultcode = util::str2int(cell->data);
                }
                else if (cell->name == "timestamp")
                {
                    rec->timestamp = util::str2time_t(cell->data);
                }
            }
            ret->push_back(rec);
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) <<
                    "Caught exception while parsing test results: " <<
                    ex.getMessage();
            continue;
        }

    }

    return ret;
}

void Database::saveDSM(boost::shared_ptr<dSM> dsm)
{
    std::ostringstream qstr;

    qstr << "INSERT OR REPLACE INTO " << Q("dsms") << " (" << Q("dsid") << ",";
    qstr << Q("hwversion") << "," << Q("armswversion") << ",";
    qstr << Q("dspswversion") << "," << Q("apiversion") << ",";
    qstr << Q("name") << "," << Q("present") << "," << Q("apartmentid") << ",";
    qstr << Q("configid") << "," << Q("countdevices") << ",";
    qstr << Q("countactivedevices") << "," Q("countinactivedevices") << ") ";

    qstr << "VALUES (" << m_sqlite3->escape(util::str(dsm->getDSID())) << ",";
    qstr << dsm->getHWVersion() << "," << dsm->getARMSwVersion() << ",";
    qstr << dsm->getDSPSwVersion() << "," << dsm->getAPIVersion() << ",";
    qstr << m_sqlite3->escape(dsm->getName()) << ",";
    qstr << (bool) dsm->getPresent() << ",";
    qstr << (int) dsm->getApartmentId() << ",";
    qstr << dsm->getConfigurationId() << "," << dsm->getDeviceCount() << ",";
    qstr << dsm->getActiveDeviceCount() << ",";
    qstr << dsm->getInactiveDeviceCount() << ");";
   
    FILE_LOG(logDEBUG4) << "saveDSM: " << qstr.str();

    m_sqlite3->exec(qstr.str());
}

boost::shared_ptr<dSM> Database::loadDSM(dsid_t dsid)
{
    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("dsms") << " WHERE ";
    qstr << Q("dsid") << " = " << m_sqlite3->escape(util::str(dsid));
    qstr << " LIMIT 1;";

    FILE_LOG(logDEBUG4) << "loadDSM: " << qstr.str();

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("dSM with dSID " + util::str(dsid) + " not found");
    }

    return createDSMFromRow(dsid, q->at(0));
}

boost::shared_ptr<dSM> Database::createDSMFromRow(dsid_t dsid,
                                     boost::shared_ptr<SQLite3::row_result> rr)
{
    boost::shared_ptr<dSM> dsm(new dSM(dsid));

    for (size_t j = 0; j < rr->size(); j++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(j);
        if (cell->name == "hwversion")
        {
            dsm->setHWVersion(util::str2uint32_t(cell->data));
        }
        else if (cell->name == "armswversion")
        {
            dsm->setARMSwVersion(util::str2uint32_t(cell->data));
        }
        else if (cell->name == "dspswversion")
        {
            dsm->setDSPSwVersion(util::str2uint32_t(cell->data));
        }
        else if (cell->name == "apiversion")
        {
            dsm->setAPIVersion(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "name")
        {
            dsm->setName(cell->data);
        }
        else if (cell->name == "present")
        {
            dsm->setPresent((uint8_t)util::str2int(cell->data));
        }
        else if (cell->name == "apartmentid")
        {
            dsm->setApartmentId((uint8_t)util::str2int(cell->data));
        }
        else if (cell->name == "configid")
        {
            dsm->setConfigurationId(util::str2uint64_t(cell->data));
        }
        else if (cell->name == "countdevices")
        {
            dsm->setDeviceCount(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "countactivedevices")
        {
            dsm->setActiveDeviceCount(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "countinactivedevices")
        {
            dsm->setInactiveDeviceCount(util::str2uint16_t(cell->data));
        }
    }

    return dsm;
}

boost::shared_ptr<std::list<dsid_t> > Database::getDSMdSIDs()
{
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

    std::ostringstream qstr;

    qstr << "SELECT " << Q("dsid") << " FROM " << Q("dsms") << ";";
    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        boost::shared_ptr<SQLite3::row_result> rr = q->at(i);
        if (rr->empty())
        {
            FILE_LOG(logWARNING) << "getDSMdSIDs: ignoring invalid row";
            continue;
        }

        boost::shared_ptr<SQLite3::cell> cell = rr->at(0);
        if (cell->name != "dsid")
        {
            FILE_LOG(logWARNING) << "getDSMdSIDs: ignoring invalid data";
            continue;
        }
        dsm_list->push_back(util::str2dsid(cell->data));
    }

    return dsm_list;    
}

void Database::deleteDSM(dsid_t dsid)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("dsms") << " WHERE " << Q("dsid") << " = ";
    qstr << m_sqlite3->escape(util::str(dsid)) << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::updateDSMPresence(bool presence)
{
    std::ostringstream qstr;
    qstr << "UPDATE " << Q("dsms") << " SET " << Q("present") << " = ";
    qstr << presence <<";";
    m_sqlite3->exec(qstr.str());
}

void Database::updatePresence(bool presence)
{
    std::ostringstream qstr;
    qstr << "BEGIN TRANSACTION;";

    qstr << "UPDATE " << Q("dsms") << " SET " << Q("present") << " = ";
    qstr << presence <<";";
    qstr << "UPDATE " << Q("devices") << " SET " << Q("state") << " = ";
    qstr << presence <<";";

    qstr << "COMMIT;";
    m_sqlite3->exec(qstr.str());
}

std::string Database::deviceKey(dsid_t dsid, dsid_t dsm_dsid)
{
    return util::str(dsid) + util::str(dsm_dsid);
}

void Database::saveDevice(boost::shared_ptr<Device> device)
{
    std::ostringstream qstr;

    qstr << "INSERT OR REPLACE INTO " << Q("devices") << " (";
    qstr << Q("key") << "," << Q("dsid") << "," << Q("dsmdsid") << ",";
    qstr << Q("deviceid") << "," << Q("productid") << ",";
    qstr << Q("functionid") << "," << Q("version") << "," << Q("zoneid") << ",";
    qstr << Q("state") << "," << Q("locked") << ",";
    qstr << Q("outputmode") << "," << Q("name") << ",";
    qstr << Q("serialnumber") << ") ";

    qstr << "VALUES (" << m_sqlite3->escape(deviceKey(device->getDSID(), 
                                                      device->getDSMdSID()));
    qstr << "," << m_sqlite3->escape(util::str(device->getDSID())) << ",";
    qstr << m_sqlite3->escape(util::str(device->getDSMdSID())) << ",";
    qstr << device->getId() << "," << device->getProductId() << ",";
    qstr << device->getFunctionId() << "," << device->getVersion() << ",";
    qstr << device->getZoneId() << "," << (bool)device->getState() << ",";
    qstr << (bool)device->getLocked() << ",";
    qstr << (bool)device->getOutputMode() << ",";
    qstr << m_sqlite3->escape(device->getName()) << ",";
    qstr << device->getSerialNumber() << ");";
   
    FILE_LOG(logDEBUG4) << "saveDevice: " << qstr.str();

    m_sqlite3->exec(qstr.str());
}

void Database::deleteDevice(dsid_t dsid)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("devices") << " WHERE " << Q("dsid") << " = ";
    qstr << m_sqlite3->escape(util::str(dsid)) << ";";
    m_sqlite3->exec(qstr.str());
}

void Database::deleteDevice(dsid_t dsid, dsid_t dsm_dsid)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("devices") << " WHERE " << Q("dsid") << " = ";
    qstr << m_sqlite3->escape(util::str(dsid)) << " AND ";
    qstr << Q("dsmdsid") << " = " << m_sqlite3->escape(util::str(dsm_dsid));
    qstr << ";";
    FILE_LOG(logDEBUG4) << "deleteDevice(dsid, dsm_dsid): " << qstr.str();
    m_sqlite3->exec(qstr.str());
}

void Database::deleteDevice(dsid_t dsid, uint16_t device_id)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("devices") << " WHERE " << Q("dsid") << " = ";
    qstr << m_sqlite3->escape(util::str(dsid)) << " AND ";
    qstr << Q("deviceid") << " = " << device_id << ";";
    FILE_LOG(logDEBUG4) << "deleteDevice(dsid, device_id): " << qstr.str();
    m_sqlite3->exec(qstr.str());
}

void Database::deleteDevice(dsid_t dsid, dsid_t dsm_dsid, uint16_t device_id)
{
    std::ostringstream qstr;
    qstr << "DELETE FROM " << Q("devices") << " WHERE " << Q("dsid") << " = ";
    qstr << m_sqlite3->escape(util::str(dsid)) << " AND ";
    qstr << Q("dsmdsid") << " = " << m_sqlite3->escape(util::str(dsm_dsid));
    qstr << " AND " << Q("deviceid") << " = " << device_id << ";";
    FILE_LOG(logDEBUG4) << "deleteDevice(dsid, dsm_dsid, device_id): "
                        << qstr.str();
    m_sqlite3->exec(qstr.str());
}

boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                        Database::loadDevices(dsid_t dsm_dsid, bool only_active)
{
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > dev_list(
            new std::list<boost::shared_ptr<Device> >);

    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("devices") << "  WHERE ";
    qstr << Q("dsmdsid") << " = " << m_sqlite3->escape(util::str(dsm_dsid));
    if (only_active)
    {
        qstr << " AND " << Q("state") << " = 1";
    }
    qstr << ";";
    FILE_LOG(logDEBUG4) << "load devices: " << qstr.str();

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        boost::shared_ptr<SQLite3::row_result> rr = q->at(i);
        if (rr->empty())
        {
            FILE_LOG(logWARNING) << "loadDevices: ignoring empty row";
            continue;
        }

        try
        {
            boost::shared_ptr<Device> d = createDeviceFromRow(q->at(i));
            dev_list->push_back(d);
        }
        catch (Exception ex)
        {
            FILE_LOG(logWARNING) << "loadDevices: " << ex.getMessage();
            continue;
        }
    }

    return dev_list;
}

boost::shared_ptr<std::list<boost::shared_ptr<Device> > >
                                        Database::loadDevices(bool only_active)
{
    boost::shared_ptr<std::list<boost::shared_ptr<Device> > > dev_list(
            new std::list<boost::shared_ptr<Device> >);

    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("devices");
    if (only_active)
    {
        qstr << " WHERE " << Q("state") << " = 1";
    }
    qstr << ";";
    FILE_LOG(logDEBUG4) << "load devices: " << qstr.str();

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());

    for (size_t i = 0; i < q->size(); i++)
    {
        boost::shared_ptr<SQLite3::row_result> rr = q->at(i);
        if (rr->empty())
        {
            FILE_LOG(logWARNING) << "loadDevices: ignoring empty row";
            continue;
        }

        try
        {
            boost::shared_ptr<Device> d = createDeviceFromRow(q->at(i));
            dev_list->push_back(d);
        }
        catch (Exception ex)
        {
            FILE_LOG(logWARNING) << "loadDevices: " << ex.getMessage();
            continue;
        }
    }

    return dev_list;
}
boost::shared_ptr<Device> Database::createDeviceFromRow(
                                    boost::shared_ptr<SQLite3::row_result> rr)
{
    boost::shared_ptr<Device> device(new Device());

    for (size_t j = 0; j < rr->size(); j++)
    {
        boost::shared_ptr<SQLite3::cell> cell = rr->at(j);

        if (cell->name == "dsid")
        {
            device->setDSID(util::str2dsid(cell->data));
        }
        else if (cell->name == "dsmdsid")
        {
            device->setDSMdSID(util::str2dsid(cell->data));
        }
        else if (cell->name == "deviceid")
        {
            device->setId(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "productid")
        {
            device->setProductId(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "functionid")
        {
            device->setFunctionId(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "version")
        {
            device->setVersion(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "zoneid")
        {
            device->setZoneId(util::str2uint16_t(cell->data));
        }
        else if (cell->name == "state")
        {
            device->setState((uint8_t)util::str2int(cell->data));
        }
        else if (cell->name == "locked")
        {
            device->setLocked((uint8_t)util::str2int(cell->data));
        }
        else if (cell->name == "outputmode")
        {
            device->setOutputMode((uint8_t)util::str2int(cell->data));
        }
        else if (cell->name == "name")
        {
            device->setName(cell->data);
        }
        else if (cell->name == "serialnumber")
        {
            device->setSerialNumber(util::str2uint32_t(cell->data));
        }
    }

    return device;
}

boost::shared_ptr<Device> Database::loadDevice(dsid_t dsid)
{
    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("devices") << " WHERE ";
    qstr << Q("dsid") << " = " << m_sqlite3->escape(util::str(dsid));
    qstr << " AND " << Q("state") << " = \"1\";";

    FILE_LOG(logDEBUG4) << "loadDevice: " << qstr.str();

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("Device with dSID " + util::str(dsid) + " not found");
    }
    if (q->size() > 1)
    {
        throw _Exception("Multiple active devices with dSID " +
                util::str(dsid) + " found");
    }

    return createDeviceFromRow(q->at(0));
}

boost::shared_ptr<Device> Database::loadDevice(dsid_t dsid,
                                                     dsid_t dsm_dsid)
{
    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("devices") << " WHERE ";
    qstr << Q("dsid") << " = " << m_sqlite3->escape(util::str(dsid));
    qstr << " AND " << Q("dsmdsid") << " = ";
    qstr << m_sqlite3->escape(util::str(dsm_dsid)) << " LIMIT 1;";

    FILE_LOG(logDEBUG4) << "loadDevice: " << qstr.str();

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    if (q->size() == 0)
    {
        throw _Exception("Device with dSID " + util::str(dsid) + " on dSM " + 
                         util::str(dsm_dsid) + " not found");
    }

    return createDeviceFromRow(q->at(0));
}

boost::shared_ptr<Device> Database::loadDevice(dsid_t dsm_dsid,
                                                     uint16_t device_id)
{
    std::ostringstream qstr;

    qstr << "SELECT * FROM " << Q("devices") << " WHERE ";
    qstr << Q("dsmdsid") << " = " << m_sqlite3->escape(util::str(dsm_dsid));
    qstr << " AND " << Q("deviceid") << " = " << device_id;
    qstr << " AND " << Q("state") << " = \"1\";";

    FILE_LOG(logDEBUG4) << "loadDevice: " << qstr.str();

    boost::shared_ptr<SQLite3::query_result> q = m_sqlite3->query(qstr.str());
    size_t num_results = q->size();
    if (num_results == 0)
    {
        throw _Exception("Device with id " + util::str(device_id) + " on dSM " +
                         util::str(dsm_dsid) + " not found");
    }
    else if (num_results > 1)
    {
        FILE_LOG(logWARNING) << "Database: found " << util::str(num_results) <<
                " active devices with DeviceId " << util::str(device_id) <<
                " on dSM " << util::str(dsm_dsid);
    }

    return createDeviceFromRow(q->at(num_results - 1));
}


Database::~Database()
{
    if (m_sqlite3)
    {
        delete m_sqlite3;
    }
}

} // namespace
